/*
 * Name:    mx_simple_server.c
 *
 * Purpose: mx_simple_server is a simplified interface to an MX control
 *          system.  It is primarily intended to make it easy for commercial
 *          vendor's software to interface with MX, but it may be useful for
 *          other purposes as well.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "mx_record.h"
#include "mx_driver.h"
#include "mx_socket.h"
#include "mx_select.h"
#include "mx_version.h"
#include "mx_variable.h"
#include "mx_analog_input.h"
#include "mx_motor.h"

/* Definitions and function prototypes. */

#ifdef OS_WIN32
    extern char *optarg;
    extern int optind;
#else
#   include <getopt.h>
#endif

void
mxss_output_function( char *string )
{
	/* Currently we throw error, debug, etc. messages away. */

	return;
}

mx_status_type
mxss_setup_server( int port_number, MX_SOCKET *socket )
{
	mx_status_type mx_status;

	if ( port_number < 0 ) {
		setvbuf(stdin,  (char *) NULL, _IOLBF, 0);
		setvbuf(stdout, (char *) NULL, _IOLBF, 0);

		*socket = -1;

		return MX_SUCCESSFUL_RESULT;
	} else {
		mx_status = mx_socket_open_as_server( socket, port_number, 0 );

		return mx_status;
	}
}

mx_status_type
mxss_shutdown_server( MX_SOCKET socket )
{
	mx_status_type mx_status;

	if ( socket == MX_INVALID_SOCKET ) {
		fclose( stdout );

		return MX_SUCCESSFUL_RESULT;
	} else {
		mx_status = mx_socket_close( socket );

		return mx_status;
	}
}

mx_status_type
mxss_socket_wait_for_input( MX_SOCKET mx_socket )
{
	static const char fname[] = "mxss_socket_wait_for_input()";

	fd_set readfds;
	struct timeval timeout;
	int num_fds, max_fd, saved_errno;

	/* Loop waiting for input. */

	while (1) {
		/* Initialize the arguments to select(). */

		FD_ZERO( &readfds );

		FD_SET( mx_socket, &readfds );

#ifdef OS_WIN32
		max_fd = -1;
#else
		max_fd = mx_socket + 1;
#endif
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;		/* Wait one second */

		num_fds = select( max_fd, &readfds, NULL, NULL, &timeout );

		saved_errno = errno;

		if ( num_fds < 0 ) {
			return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Error in select() while waiting for events.  "
			"Error = %d.  Error string = '%s'.",
			saved_errno, strerror( saved_errno ) );

		} else if ( num_fds == 0 ) {

			/* We timed out, so go back to the top of the loop. */

		} else {
			if ( FD_ISSET( mx_socket, &readfds ) == 0 ) {
				return mx_error( MXE_NETWORK_IO_ERROR, fname,
	"select() reported activity, but not on the expected socket %d.",
					mx_socket );
			}

			/* Data is available on the socket so return. */

			return MX_SUCCESSFUL_RESULT;
		}
	}
}

mx_status_type
mxss_socket_wait_for_connection( MX_SOCKET server_socket,
				MX_SOCKET *client_socket )
{
	static const char fname[] = "mxss_socket_wait_for_connection()";

	struct sockaddr_in client_address;
	int client_address_size = sizeof( struct sockaddr_in );
	int saved_errno;
	char *error_string;
	mx_status_type mx_status;

	/* If we are using a pipe rather than a socket, return immediately. */

	if ( server_socket == MX_INVALID_SOCKET ) {
		*client_socket = MX_INVALID_SOCKET;

		return MX_SUCCESSFUL_RESULT;
	}

	/* Wait for activity on the socket. */

	mx_status = mxss_socket_wait_for_input( server_socket );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* The server socket should be ready for an accept() operation,
	 * so let's try it.
	 */

	*client_socket = accept( server_socket,
				( struct sockaddr *) &client_address,
				&client_address_size );

        saved_errno = mx_socket_check_error_status(
                        client_socket, MXF_SOCKCHK_INVALID, &error_string );

        if ( saved_errno != 0 ) {
                return mx_error( MXE_NETWORK_IO_ERROR, fname,
               "Attempt to perform accept() on server port failed.  "
                        "Errno = %d.  Error string = '%s'.",
                        saved_errno, error_string );
        }

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxss_get_command_line( MX_SOCKET socket,
			char *buffer,
			size_t buffer_length )
{
	size_t length;
	mx_status_type mx_status;

	if ( socket == MX_INVALID_SOCKET ) {
		fgets( buffer, buffer_length, stdin );

		/* Zap the newline at the end of the buffer. */

		length = strlen( buffer );

		if ( buffer[length-1] == '\n' ) {
			buffer[length-1] = '\0';
		}

		return MX_SUCCESSFUL_RESULT;
	} else {
		/* Wait for a command to arrive. */

		mx_status = mxss_socket_wait_for_input( socket );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* Wait a while for the command to finish arriving. */

		mx_msleep(100);		/* 100 milliseconds */

		/* Read the command. */

		mx_status = mx_socket_getline( socket,
					buffer, buffer_length );
		return mx_status;
	}
}

mx_status_type
mxss_send_response( MX_SOCKET socket,
			char *command_name,
			mx_status_type command_status,
			char *response_string )
{
	char response[500];
	mx_status_type mx_status;

	if ( command_status.code == MXE_SUCCESS ) {
		if ( response_string != NULL ) {
			sprintf( response, "%s ok %s",
				command_name, response_string );
		} else {
			sprintf( response, "%s ok",
				command_name );
		}
	} else {
		sprintf( response, "%s error %s %s", command_name,
				mx_strerror( command_status.code, NULL, 0 ),
				command_status.message );
	}

	if ( socket == MX_INVALID_SOCKET ) {
		fprintf( stdout, "%s\n", response );

		return MX_SUCCESSFUL_RESULT;
	} else {
		mx_status = mx_socket_putline( socket, response );

		return mx_status;
	}
}

/******* Server main routine. *******/

int
main( int argc, char *argv[] )
{
	static const char pname[] = "mx_simple_server";

	MX_SOCKET server_socket, client_socket;
	MX_RECORD *record_list, *record;
	char buffer[200];
	char response[2000];
	char database_name[ MXU_FILENAME_LENGTH + 1 ];
	char record_name[ MXU_RECORD_NAME_LENGTH + 1 ];
	double position, destination, value;
	int c, num_items, error_flag, port_number;
	mx_status_type mx_status;

	static char usage[] = 
		"Usage: %s [ -f database_filename ] [ -p server_port ]\n";

	mx_set_error_output_function( mxss_output_function );

	/* Construct the default database filename */

	strcpy( database_name, "/opt/mx/etc/motor.dat" );

	/* Process command line arguments, if any. */

	port_number = -1;
	error_flag = FALSE;

	while ((c = getopt(argc, argv, "f:p:")) != -1 ) {
		switch (c) {
		case 'f':
			mx_strncpy( database_name, optarg,
					MXU_FILENAME_LENGTH );
			break;
		case 'p':
			port_number = atoi( optarg );
			break;
		case '?':
			error_flag = TRUE;
			break;
		}
	}

	if ( error_flag == TRUE ) {
		fprintf( stderr, usage );
		exit(1);
	}

	/* Initialize the MX database. */

	mx_status = mx_setup_database( &record_list, database_name );

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	/* Initialize the server. */

	mx_status = mxss_setup_server( port_number, &server_socket );

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	/* Wait for a client connection. */

	mx_status = mxss_socket_wait_for_connection(
				server_socket, &client_socket );

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	/* Enter the server event loop. */

	while( 1 ) {

		mx_status = mxss_get_command_line( client_socket,
						buffer, sizeof buffer );

		if ( mx_status.code != MXE_SUCCESS )
			exit(1);

		/* Execute the command. */

		if ( strncmp( buffer, "quit", 4 ) == 0 ) {

			mxss_shutdown_server( client_socket );

			exit(0);

		} else if ( strncmp( buffer, "exit", 4 ) == 0 ) {

			mxss_shutdown_server( client_socket );

			exit(0);

		} else if ( strncmp( buffer, "version", 7 ) == 0 ) {

			sprintf( response, "MX %s", mx_get_version_string() );

			mx_status = mxss_send_response( client_socket,
				"version", MX_SUCCESSFUL_RESULT, response );

			if ( mx_status.code != MXE_SUCCESS )
				exit(1);

		} else if ( strncmp( buffer, "move", 4 ) == 0 ) {

			num_items = sscanf( buffer, "%*s %s %lg",
						record_name, &destination );

			if ( num_items != 2 ) {

				mx_status = mxss_send_response( client_socket,
					"move",
					mx_error( MXE_UNPARSEABLE_STRING, pname,
				"Badly formatted move command '%s'",
					buffer ), NULL );

				if ( mx_status.code != MXE_SUCCESS )
					exit(1);

				/* Go back to the top of the while loop. */

				continue;
			}

			record = mx_get_record( record_list, record_name );

			if ( record == NULL ) {

				mx_status = mxss_send_response( client_socket,
					"move",
					mx_error( MXE_NOT_FOUND, pname,
				"The motor record '%s' does not exist.",
						record_name ), NULL );

				if ( mx_status.code != MXE_SUCCESS )
					exit(1);

				/* Go back to the top of the while loop. */

				continue;
			}

			mx_status = mx_motor_move_absolute( record,
							destination, 0 );

			if ( mx_status.code != MXE_SUCCESS ) {

				mx_status = mxss_send_response( client_socket,
						"move", mx_status, NULL );

				if ( mx_status.code != MXE_SUCCESS )
					exit(1);

				/* Go back to the top of the while loop. */

				continue;
			}

			mx_status = mx_wait_for_motor_stop( record, 0 );

			mx_status = mxss_send_response( client_socket,
						"move", mx_status, NULL );

			if ( mx_status.code != MXE_SUCCESS )
				exit(1);

		} else if ( strncmp( buffer, "get_position", 12 ) == 0 ) {

			num_items = sscanf( buffer, "%*s %s", record_name );

			if ( num_items != 1 ) {

				mx_status = mxss_send_response( client_socket,
					"get_position",
					mx_error( MXE_UNPARSEABLE_STRING, pname,
				"Badly formatted get_position command '%s'",
					buffer ), NULL );

				if ( mx_status.code != MXE_SUCCESS )
					exit(1);

				/* Go back to the top of the while loop. */

				continue;
			}

			record = mx_get_record( record_list, record_name );

			if ( record == NULL ) {

				mx_status = mxss_send_response(
					client_socket, "get_position",
					mx_error( MXE_NOT_FOUND, pname,
				"The motor record '%s' does not exist.",
						record_name ), NULL );

				if ( mx_status.code != MXE_SUCCESS )
					exit(1);

				/* Go back to the top of the while loop. */

				continue;
			}

			mx_status = mx_motor_get_position( record, &position );

			if ( mx_status.code == MXE_SUCCESS ) {
				sprintf( response,"%g", position );

				mx_status = mxss_send_response( client_socket,
					"get_position", mx_status, response );
			} else {
				mx_status = mxss_send_response( client_socket,
					"get_position", mx_status, NULL );
			}

		} else if ( strncmp( buffer, "get_value", 9 ) == 0 ) {

			num_items = sscanf( buffer, "%*s %s", record_name );

			if ( num_items != 1 ) {

				mx_status = mxss_send_response( client_socket,
					"get_value",
					mx_error( MXE_UNPARSEABLE_STRING, pname,
				"Badly formatted get_value command '%s'",
					buffer ), NULL );

				if ( mx_status.code != MXE_SUCCESS )
					exit(1);

				/* Go back to the top of the while loop. */

				continue;
			}

			record = mx_get_record( record_list, record_name );

			if ( record == NULL ) {

				mx_status = mxss_send_response(
					client_socket, "get_value",
					mx_error( MXE_NOT_FOUND, pname,
					"The record '%s' does not exist.",
						record_name ), NULL );

				if ( mx_status.code != MXE_SUCCESS )
					exit(1);

				/* Go back to the top of the while loop. */

				continue;
			}

			if ( record->mx_superclass == MXR_VARIABLE ) {

				mx_status = mx_get_double_variable( record,
								&value );

			} else if ( record->mx_class == MXC_ANALOG_INPUT ) {

				mx_status = mx_analog_input_read( record,
								&value );
			} else {
				mx_status = mxss_send_response(
					client_socket, "get_value",
					mx_error( MXE_UNSUPPORTED, pname,
			"get_value is not currently supported for record '%s'.",
						record_name ), NULL );

				if ( mx_status.code != MXE_SUCCESS )
					exit(1);

				/* Go back to the top of the while loop. */

				continue;
			}

			if ( mx_status.code == MXE_SUCCESS ) {
				sprintf( response,"%g", value );

				mx_status = mxss_send_response( client_socket,
					"get_value", mx_status, response );
			} else {
				mx_status = mxss_send_response( client_socket,
					"get_value", mx_status, NULL );
			}


		} else {
			mx_status = mxss_send_response( client_socket, "cmd", 
					mx_error( MXE_ILLEGAL_ARGUMENT, pname,
					"Unrecognized command '%s'", buffer ),
					NULL );

			if ( mx_status.code != MXE_SUCCESS )
				exit(1);
		}
	}

	fprintf(stderr, "Error: Connection with client lost, shutting down.\n");

	exit(0);
}

